/*-----------------------------------------------------------------------------

 2002, Jeff M Soft und Hardware GmbH, All Rights Reserved

-----------------------------------------------------------------------------*/

#include "Module.h"
#include "SEMod_struct.h"
#include "SEPin.h"
#include <assert.h>

#define PN_INPUT1 0
#define PN_MIDI_OUT 2

#define SYSTEM_MSG		        0xF0
#define NOTE_OFF                0x80
#define NOTE_ON                 0x90

// The core of a MIDI plugin
void Module::OnMidiData( unsigned long p_clock, unsigned long midi_msg, short pin_id)
{
	// system messages are for timing info etc
	// they don't have a channel, just pass them right through
	bool is_system_msg = ( midi_msg & SYSTEM_MSG ) == SYSTEM_MSG;

	if( !is_system_msg )
	{
		// The first byte of a midi message contains the status (note on/off etc) and the channel
		int chan = midi_msg & 0x0f;
		int stat = midi_msg & 0xf0;

		// the next two bytes depend on the type of message
		int b2 = ( midi_msg / 0x100 ) & 0xff;
		int b3 = ( midi_msg / 0x10000 ) & 0xff;
		int thishigh;
		int thislow;
		switch( stat )
		{
		case NOTE_ON:
			cache[b2][0] = 1;
			cache[b2][1] = b3;
			switch(mode){
			case 1:			
				thishigh=getHigh();
				if(thishigh!=0){
					if(highest < thishigh){
						// send old note off
						if(highest != 0){
							midi_msg = NOTE_OFF + chan + (highest << 8) + (0 << 16);
							getPin(1)->TransmitMIDI( p_clock, midi_msg );
						}
						// send new note on
						midi_msg = stat + chan + (thishigh << 8) + (cache[thishigh][1] << 16);
						getPin(1)->TransmitMIDI( p_clock+5, midi_msg );
						
						// reset highest
						highest = thishigh;
					}
				}
				break;
			case 2:
				thislow=getLow();
				//midi_msg = stat + chan + (thislow << 8) + (cache[thislow][1] << 16);
				//getPin(1)->TransmitMIDI( p_clock, midi_msg );
				if(thislow != 0){
					if(lowest > thislow){
						if(lowest != 150){
							// send old note off
							midi_msg = NOTE_OFF + chan + (lowest << 8) + (0 << 16);
							getPin(1)->TransmitMIDI( p_clock, midi_msg );
						}
						// send new note on
						midi_msg = stat + chan + (thislow << 8) + (cache[thislow][1] << 16);
						getPin(1)->TransmitMIDI( p_clock+5, midi_msg );
						// reset lowest
						lowest = thislow;
					}
				}else{
						midi_msg = stat + chan + (99 << 8) + (99 << 16);
						getPin(1)->TransmitMIDI( p_clock+5, midi_msg );
				
				}
			break;
			}

		break;
		case NOTE_OFF:
			cache[b2][0] = 0;
			cache[b2][1] = 0;
			switch(mode){
				case 1:
						// kill current highest note ...
					if(b2 == highest  ){
						// kill old highest

						midi_msg = stat + chan + (highest << 8) + (0 << 16);
						getPin(1)->TransmitMIDI( p_clock+0, midi_msg );

						// send new highest
						thishigh=getHigh();
						if(thishigh != 0){
							midi_msg = NOTE_ON + chan + (thishigh << 8) + (cache[thishigh][1] << 16);
							getPin(1)->TransmitMIDI( p_clock+5, midi_msg );
						}
						highest = thishigh;
					}					
				break;

				case 2:
						// kill current lowest note ...
					if(b2 == lowest  ){
						// kill old lowest
						if(lowest != 0){
						midi_msg = stat + chan + (lowest << 8) + (0 << 16);
						getPin(1)->TransmitMIDI( p_clock+0, midi_msg );
						};

						// send new lowest
						thislow=getLow();
						if(thislow != 150 && thislow != 0){
							midi_msg = NOTE_ON + chan + (thislow << 8) + (cache[thislow][1] << 16);
							getPin(1)->TransmitMIDI( p_clock+5, midi_msg );
							lowest = thislow;
						}else{
							lowest=150;
						};
						
						
					}					

				break;
			}

			// re-assemble the MIDI message into an 32 bit integer
			// midi_msg = stat + chan + (b2 << 8) + (b3 << 16);
			break;
		}
	}

	// send the MIDI
	//getPin(1)->TransmitMIDI( p_clock, midi_msg );
}

int Module::getHigh(){
	int high=0;
	for(int i=0;i<200;i++){
		if(cache[i][0] == 1){
			high=i;
		};
	};
	return high;
}
int Module::getLow(){
	int low=0;
	for(int i=0;i<200;i++){
		if(cache[i][0] == 1){
			if(low == 0){
				low=i;
			};
		};
	};
	return low;
}
Module::Module(seaudioMasterCallback2 seaudioMaster, void *p_resvd1) : SEModule_base(seaudioMaster, p_resvd1)
{
}

void Module::open()
{
	SEModule_base::open();
	// choose which function is used to process audio
	SET_PROCESS_FUNC(Module::sub_process);
	// clear note cache ready to start
	for(int i=0;i<200;i++){
		cache[i][0]=cache[i][1]=0;
	}
	lowest = 150;
	highest = 0;
}

Module::~Module()
{
	// nothing to see here, move along.
}

// describe the module
bool Module::getModuleProperties (SEModuleProperties* properties)
{
	properties->name = "dd_MIDIHighLow";	// this is the name the end-user will see
	properties->id = "dd_MIDIHighLow";
	properties->about = "http://www.roughdiamondproductions.com";
	return true;
}

// describe the pins (plugs)
bool Module::getPinProperties (long index, SEPinProperties* properties)
{
	switch( index )
	{
	case 0:
		properties->name				= "MIDI In";
		properties->direction			= DR_IN;
		properties->datatype			= DT_MIDI2;
		break;
	case 1:
		properties->name				= "MIDI Out";
		properties->direction			= DR_OUT;
		properties->datatype			= DT_MIDI2;
		break;
	case 2:
		properties->name				= "Mode";
		properties->variable_address	= &mode;
		properties->datatype_extra		= "Lowest=2,Highest=1";
		properties->default_value		= "1";
		properties->direction			= DR_IN;
		properties->datatype			= DT_ENUM;
		break;
	default:
		return false; // host will ask for plugs 0,1,2,3 etc. return false to signal when done
	};

	return true;
}

void Module::sub_process(long buffer_offset, long sampleFrames )
{
	// Nothing to do here in a pure MIDI plugin
//	if( !AreEventsPending() )
	{
		CallHost(seaudioMasterSleepMode);
	}
}

void Module::OnPlugStateChange(SEPin *pin)
{
	// Nothing to do here in a pure MIDI plugin
}